import os
import copy
import math
import numpy as np
import matplotlib.pyplot as plt

from scipy import stats
from ...utils.quant_utils import float2fixed


class QuantBase:
    def __init__(self, name, module, bit_width, logger, writer):
        self.name = name
        self.module = module
        self.bit_width = bit_width
        self.logger = logger

        self.weight_fl = 0
        self.activation_fl = 0
        self.activation_kld = None

        self.activation = None
        self.activation_positive = True
        self.bins_num = 2**(2 * self.bit_width - 1)
        self.activation_max = 2**(bit_width - 1) - 1
        self.activation_distribution = np.zeros(self.bins_num)
        self.activation_distribution_edge = None

        # if self.module have hook, self.hook_handler.remove() will remove it.
        self.hook_handler = None
        # assign this to neutral num to decide concat.
        self.concat_sign = -1
        # True for bn2scale
        self.scale_sign = False

        self.writer = writer
        # for better fig
        self.step_count = 0

    def register_get_activation_hook(self):
        def hook(module, input, output):
            self.activation = output.cpu().detach().data.flatten().numpy()

        self.hook_handler = self.module.register_forward_hook(hook)

    def register_eval_hook(self, debug_bw=None, debug_fl=None):
        def hook(module, input, output):
            bit_width = self.bit_width if debug_bw is None else debug_bw
            fraction_length = self.activation_fl if debug_fl is None else debug_fl
            output = float2fixed(output, bit_width, fraction_length)
            return output

        if self.activation_fl is not -1:
            self.hook_handler = self.module.register_forward_hook(hook)

    def init_histograms(self):
        hist, hist_edge = np.histogram(self.activation,
                                       bins=self.bins_num,
                                       range=(-1 * self.activation_max,
                                              self.activation_max))
        self.activation_distribution += hist
        self.activation_distribution_edge = hist_edge

        self.step_count += 1
        self.writer.add_histogram(self.name + "_running_activation",
                                  self.activation, self.step_count)

    def reweight_kld(self):
        def adaptive_func(x):
            x /= 128 * 128
            return abs(x)

        adaptive_scale = [
            adaptive_func(x - self.bins_num // 2) for x in range(self.bins_num)
        ]
        self.activation_distribution *= adaptive_scale

    def adaptive_kld(self):
        # -1, 1
        adaptive_zone = copy.deepcopy(
            self.activation_distribution[self.bins_num // 2 -
                                         2**(self.bit_width -
                                             1):self.bins_num // 2 +
                                         2**(self.bit_width - 1)])
        activation_sum = sum(self.activation_distribution)
        adaptive_zero = 2**(self.bit_width - 1)
        # handcraft rule
        alpha = 0
        beta = 2**(2 - self.bit_width) / 8
        weight = 0
        for i in range(self.bit_width - 1, -1, -1):
            weight = sum(adaptive_zone[adaptive_zero - 2**
                                       (self.bit_width - 1 - i):adaptive_zero +
                                       2**(self.bit_width - 1 - i)]) - weight
            # handcraft rule
            if weight / activation_sum > 0.1:
                beta *= 2
            # handcraft rule
            alpha += (2**(i + 1)) * (weight / activation_sum)

        def adaptive_func(alpha, beta, x):
            # 防止计算指数时溢出
            x /= 128 * 8
            # 防止计算指数时溢出
            if x > 0:
                param = min(20, max(-20, -1 * alpha * (-1 * x + beta)))
                return 1 - 1 / (1 + (math.exp(param)))
            else:
                param = min(20, max(-20, -1 * alpha * (x + beta)))
                return 1 - 1 / (1 + (math.exp(param)))

        adaptive_scale = [
            adaptive_func(alpha, beta, x - self.bins_num // 2)
            for x in range(self.bins_num)
        ]

        self.activation_distribution *= adaptive_scale

    def quantize_params(self):
        # backup data, need overide
        self.backup_weight = copy.deepcopy(self.module.weight)
        self.writer.add_histogram(
            self.name + "_weight",
            self.backup_weight.clone().cpu().data.numpy(),
        )
        weight_data = self.module.weight.cpu().data.flatten().numpy()
        try:
            # many module may not have bias
            bias_data = self.module.bias.cpu().data.flatten().numpy()
            # backup bias
            self.backup_bias = self.module.bias
            self.writer.add_histogram(
                self.name + "_bias",
                self.backup_bias.clone().cpu().data.numpy(),
            )
        except:
            bias_data = None
            # no bias backup
            self.backup_bias = None
        if bias_data is None:
            data = weight_data
        else:
            data = np.concatenate((weight_data, bias_data))

        # init kld
        hist, hist_edge = np.histogram(data,
                                       bins=self.bins_num,
                                       range=(-1 * self.activation_max,
                                              self.activation_max))
        min_kld_fl, kld = kld_based_dynamics_double(hist, self.bit_width)
        self.weight_fl = min_kld_fl
        self.logger.debug(self.name + 's\' weight fl:' + str(min_kld_fl) +
                          ', kld: ' + str(kld))

    def quantize_activation(self):
        self.writer.add_histogram(self.name + "_activation",
                                  self.activation_distribution)
        # pick fl which minimizes KL divergence
        if self.activation_positive:
            self.activation_fl, self.activation_kld = kld_based_dynamics_single(
                self.activation_distribution, self.bit_width)
        else:
            # self.adaptive_kld()
            self.reweight_kld()
            self.activation_fl, self.activation_kld = kld_based_dynamics_double(
                self.activation_distribution, self.bit_width)

        self.logger.debug(self.name + 's\' activation fl:' +
                          str(self.activation_fl) + ', kld: ' +
                          str(self.activation_kld))

    def trim_params(self, debug_bw=None, debug_fl=None):
        # for debug
        bit_width = self.bit_width if not debug_bw else debug_bw
        fraction_length = self.weight_fl if not debug_fl else debug_fl

        # trim backup data
        if self.weight_fl != -1:
            if self.backup_weight is not None:
                self.module.weight.data = float2fixed(self.backup_weight,
                                                      bit_width,
                                                      fraction_length)
            if self.backup_bias is not None:
                self.module.bias.data = float2fixed(self.backup_bias,
                                                    bit_width, fraction_length)

    def recover_params(self):
        # recover backup data
        if self.weight_fl != -1:
            self.module.weight.data = self.backup_weight
            if self.backup_bias is not None:
                self.module.bias.data = self.backup_bias

    def trim_activation(self, debug_bw=None, debug_fl=None):
        # remove origin get_activation hook or other fl's hook
        if self.hook_handler is not None:
            self.hook_handler.remove()

        # rigister new fl's hook for trim activation
        self.register_eval_hook(debug_bw, debug_fl)


def kld_based_dynamics_single(distribution, bit_width):
    kld = np.zeros(bit_width)
    # only calculate positive
    zero_bin = 2**(2 * bit_width - 2)
    distribution = distribution[zero_bin + 1:]

    for i in range(bit_width):
        range_bin = 2**(i + bit_width - 2)
        p = copy.deepcopy(distribution[:range_bin]).astype(np.float64)
        q = copy.deepcopy(distribution[:range_bin]).astype(np.float64)
        # p[0] += sum(distribution[:zero_bin - range_bin])
        p[-1] += sum(distribution[range_bin:])
        is_nonzeros = (p != 0).astype(np.int64)
        #extended q
        if i != 0:
            for j in range(0, 2**(i + bit_width - 1), 2**(i - 1)):
                step = 2**(i - 1)
                norm = sum(is_nonzeros[j:j + step])
                q[j:j + step] = sum(q[j:j + step]) / norm if norm != 0 else 0
        q[p == 0] = 0
        p[p == 0] = 0.000001
        q[q == 0] = 0.000001
        # calculate kl_divergence between q and p
        kld[i] = stats.entropy(p, q)

    kld = list(kld)
    kld.reverse()
    kld = np.array(kld)

    min_kld_fl = np.argmin(kld)

    return min_kld_fl, kld


def kld_based_dynamics_double(distribution, bit_width):
    kld = np.zeros(bit_width)
    zero_bin = 2**(2 * bit_width - 2)

    for i in range(bit_width):
        range_bin = 2**(i + bit_width - 1)
        p = copy.deepcopy(distribution[zero_bin - range_bin:zero_bin +
                                       range_bin]).astype(np.float64)
        q = copy.deepcopy(distribution[zero_bin - range_bin:zero_bin +
                                       range_bin]).astype(np.float64)
        p[0] += sum(distribution[:zero_bin - range_bin])
        p[-1] += sum(distribution[zero_bin + range_bin:])
        is_nonzeros = (p != 0).astype(np.int64)
        #extended q
        if i != 0:
            for j in range(0, 2**(i + bit_width), 2**i):
                if j == 0 or j == 2**(i + bit_width) - 2**i:
                    step = 2**(i - 1)
                else:
                    step = 2**i
                norm = sum(is_nonzeros[j:j + step])
                q[j:j + step] = sum(q[j:j + step]) / norm if norm != 0 else 0
        q[p == 0] = 0
        p[p == 0] = 0.000001
        q[q == 0] = 0.000001
        # calculate kl_divergence between q and p
        kld[i] = stats.entropy(p, q)

    kld = list(kld)
    kld.reverse()
    kld = np.array(kld)

    min_kld_fl = np.argmin(kld)

    return min_kld_fl, kld



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing

import torch.nn as nn
import numpy as np

from .torchfunc import *


class QNQYOLO(nn.Module):
    def __init__(self, anchors, num_classes, img_dim=416):
        super(QNQYOLO, self).__init__()
        self.anchors = anchors
        self.num_anchors = len(anchors)
        self.num_classes = num_classes
        self.ignore_thres = 0.5
        self.mse_loss = nn.MSELoss()
        self.bce_loss = nn.BCELoss()
        self.obj_scale = 1
        self.noobj_scale = 100
        self.metrics = {}
        self.img_dim = img_dim
        self.grid_size = 0  # grid size
        #! add by dongz
        self.x_sigmoid = TorchSigmoid()
        self.y_sigmoid = TorchSigmoid()
        self.conf_sigmoid = TorchSigmoid()
        self.cls_sigmoid = TorchSigmoid()
        self.w_exp = TorchExp()
        self.h_exp = TorchExp()
        self.dummy_pred_boxed = TorchDummy()
        self.dummy_mul = TorchDotMul()
        self.output_cat = TorchCat(self.dummy_mul, self.conf_sigmoid,
                                   self.cls_sigmoid)

    def compute_grid_offsets(self, grid_size, cuda=True):
        self.grid_size = grid_size
        g = self.grid_size
        FloatTensor = torch.cuda.FloatTensor if cuda else torch.FloatTensor
        self.stride = self.img_dim / self.grid_size
        # Calculate offsets for each grid
        self.grid_x = torch.arange(g).repeat(g, 1).view([1, 1, g,
                                                         g]).type(FloatTensor)
        self.grid_y = torch.arange(g).repeat(g,
                                             1).t().view([1, 1, g,
                                                          g]).type(FloatTensor)
        self.scaled_anchors = FloatTensor([
            (a_w / self.stride, a_h / self.stride) for a_w, a_h in self.anchors
        ])
        self.anchor_w = self.scaled_anchors[:, 0:1].view(
            (1, self.num_anchors, 1, 1))
        self.anchor_h = self.scaled_anchors[:, 1:2].view(
            (1, self.num_anchors, 1, 1))

    def forward(self, x, targets=None, img_dim=None):

        # Tensors for cuda support
        FloatTensor = torch.cuda.FloatTensor if x.is_cuda else torch.FloatTensor
        LongTensor = torch.cuda.LongTensor if x.is_cuda else torch.LongTensor
        ByteTensor = torch.cuda.ByteTensor if x.is_cuda else torch.ByteTensor

        self.img_dim = img_dim
        num_samples = x.size(0)
        grid_size = x.size(2)

        prediction = (x.view(num_samples, self.num_anchors,
                             self.num_classes + 5, grid_size,
                             grid_size).permute(0, 1, 3, 4, 2).contiguous())

        # Get outputs
        #! edit by dongz
        x = self.x_sigmoid(prediction[..., 0])  # Center x
        y = self.y_sigmoid(prediction[..., 1])  # Center y
        w = prediction[..., 2]  # Width
        h = prediction[..., 3]  # Height
        pred_conf = self.conf_sigmoid(prediction[..., 4])  # Conf
        pred_cls = self.cls_sigmoid(prediction[..., 5:])  # Cls pred.

        # If grid size does not match current we compute new offsets
        if grid_size != self.grid_size:
            self.compute_grid_offsets(grid_size, cuda=x.is_cuda)

        # Add offset and scale with anchors
        pred_boxes = FloatTensor(prediction[..., :4].shape)
        pred_boxes[..., 0] = self.dummy_pred_boxed(x.data + self.grid_x)
        pred_boxes[..., 1] = self.dummy_pred_boxed(y.data + self.grid_y)
        pred_boxes[..., 2] = self.dummy_pred_boxed(
            self.w_exp(w.data) * self.anchor_w)
        pred_boxes[..., 3] = self.dummy_pred_boxed(
            self.h_exp(h.data) * self.anchor_h)

        output = self.output_cat(
            (
                self.dummy_mul(pred_boxes.view(num_samples, -1, 4),
                               self.stride),
                pred_conf.view(num_samples, -1, 1),
                pred_cls.view(num_samples, -1, self.num_classes),
            ),
            -1,
        )

        if targets is None:
            return output, 0
        else:
            iou_scores, class_mask, obj_mask, noobj_mask, tx, ty, tw, th, tcls, tconf = build_targets(
                pred_boxes=pred_boxes,
                pred_cls=pred_cls,
                target=targets,
                anchors=self.scaled_anchors,
                ignore_thres=self.ignore_thres,
            )

            obj_mask = obj_mask.bool()
            noobj_mask = noobj_mask.bool()
            # Loss : Mask outputs to ignore non-existing objects (except with conf. loss)
            loss_x = self.mse_loss(x[obj_mask], tx[obj_mask])
            loss_y = self.mse_loss(y[obj_mask], ty[obj_mask])
            loss_w = self.mse_loss(w[obj_mask], tw[obj_mask])
            loss_h = self.mse_loss(h[obj_mask], th[obj_mask])
            loss_conf_obj = self.bce_loss(pred_conf[obj_mask], tconf[obj_mask])
            loss_conf_noobj = self.bce_loss(pred_conf[noobj_mask],
                                            tconf[noobj_mask])
            loss_conf = self.obj_scale * loss_conf_obj + self.noobj_scale * loss_conf_noobj
            loss_cls = self.bce_loss(pred_cls[obj_mask], tcls[obj_mask])
            total_loss = loss_x + loss_y + loss_w + loss_h + loss_conf + loss_cls

            # Metrics
            cls_acc = 100 * class_mask[obj_mask].mean()
            conf_obj = pred_conf[obj_mask].mean()
            conf_noobj = pred_conf[noobj_mask].mean()
            conf50 = (pred_conf > 0.5).float()
            iou50 = (iou_scores > 0.5).float()
            iou75 = (iou_scores > 0.75).float()
            detected_mask = conf50 * class_mask * tconf
            precision = torch.sum(
                iou50 * detected_mask) / (conf50.sum() + 1e-16)
            recall50 = torch.sum(
                iou50 * detected_mask) / (obj_mask.sum() + 1e-16)
            recall75 = torch.sum(
                iou75 * detected_mask) / (obj_mask.sum() + 1e-16)

            self.metrics = {
                "loss": to_cpu(total_loss).item(),
                "x": to_cpu(loss_x).item(),
                "y": to_cpu(loss_y).item(),
                "w": to_cpu(loss_w).item(),
                "h": to_cpu(loss_h).item(),
                "conf": to_cpu(loss_conf).item(),
                "cls": to_cpu(loss_cls).item(),
                "cls_acc": to_cpu(cls_acc).item(),
                "recall50": to_cpu(recall50).item(),
                "recall75": to_cpu(recall75).item(),
                "precision": to_cpu(precision).item(),
                "conf_obj": to_cpu(conf_obj).item(),
                "conf_noobj": to_cpu(conf_noobj).item(),
                "grid_size": grid_size,
            }

            return output, total_loss



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from .quant_base import QuantBase

#! Deprecated
#! Dropout layer do not change activation locally

# class QuantDropout(QuantBase):
#     def __init__(self, name, module, bit_width, logger, writer):
#         super().__init__(name, module, bit_width, logger,  writer)
#         self.activation_positive = False

#     def quantize_params(self):
#         pass

dropout_switch = {
    # nn.Dropout: QuantDropout,
    # nn.Dropout2d: QuantDropout,
    # nn.Dropout3d: QuantDropout,
}



from .activation import *
from .convolution import *
from .dropout import *
from .linear import *
from .normalization import *
from .padding import *
from .pooling import *
from .recurrent import *
from .sparse import *
from .torchfunc import *
from .transformer import *
from .vision import *
from .complex import *
from .quant_base import *

qnq_switch = {
    **activation_switch,
    **convolution_switch,
    **dropout_switch,
    **linear_switch,
    **normalization_switch,
    **padding_switch,
    **pooling_switch,
    **recurrent_switch,
    **sparse_switch,
    **torchfunc_switch,
    **transformer_switch,
    **vision_switch
}


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from torch.nn.modules.conv import ConvTranspose1d
from .quant_base import QuantBase


# support dilation convolution
class QuantConv(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False


class QuantConvTranspose(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False


convolution_switch = {
    nn.Conv1d: QuantConv,
    nn.Conv2d: QuantConv,
    nn.Conv3d: QuantConv,
    nn.ConvTranspose1d: QuantConvTranspose,
    nn.ConvTranspose2d: QuantConvTranspose,
    nn.ConvTranspose3d: QuantConvTranspose
}


# This script contain functions help quantization
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing

import json
from .operators import *
from .utils import *

from tensorboardX import SummaryWriter


class QNQ():
    def __init__(self,
                 model,
                 save_path="./checkpoints/test",
                 config_file=None,
                 ext_metrics=None,
                 steper=None):
        # init parameters
        self.model = model
        self.save_path = save_path
        self.config_file = config_file
        # ext_metrics return a float value for search
        self.ext_metrics = ext_metrics
        self.steper = steper

        # init abstract log path and logger
        self.logger, self.log_path = lognow(self.save_path)
        self.writer = SummaryWriter(log_dir=os.path.normpath(self.log_path +
                                                             '/' + 'tblogs'))

        # load custom config
        try:
            self.load_config()
        except IOError:
            if not os.path.exists(os.path.dirname(self.config_file)):
                os.makedirs(os.path.dirname(self.config_file))
            self.init_config()
            self.logger.warning("Detect no config, creating...")
            self.logger.warning("Please review config and re-run.")
            exit(1)

        # processing custom config
        # exec fuse
        if self.fuse_all:
            fuse_module(self.model)
        else:
            self.fuse(self.config_file)

        # initialize quantization
        self.quant_layer_list = []
        self.quant_names = []
        self.logger.info("Initialize quant layer.")
        self.logger.info("Initialize activation hook.")
        for name, module in self.model.named_modules():
            quant_layer = qnq_switch.get(type(module), None)
            if quant_layer:
                # initialize quant_layer
                quant_layer = quant_layer(name, module,
                                          self.qconfig[name]['bit_width'],
                                          self.logger, self.writer)
                quant_layer.weight_fl = self.qconfig[name]['weight_fl']
                quant_layer.activation_fl = self.qconfig[name]['activation_fl']
                quant_layer.concat_sign = self.qconfig[name]['concat_sign']
                quant_layer.scale_sign = self.qconfig[name]['scale_sign']
                # append quant_layer and name
                self.quant_layer_list.append(quant_layer)
                self.quant_names.append(name)
                quant_layer.register_get_activation_hook()
        self.transfer_bn2scale()
        self.logger.info("Here the quant layers.")
        self.logger.info(str(self.quant_names))

    @exe_time
    def search(self):
        #todo when 2 or more fl have same metrics, how to do?
        def quantize_params():
            self.logger.info("Quantizing params...")
            for layer in self.quant_layer_list:
                layer.quantize_params()

        def quantize_activation():
            self.logger.info("Quantizing activation...")
            for layer in self.quant_layer_list:
                layer.quantize_activation()

        def search_params(layer, metrics_tmp):
            # info log
            self.logger.info("      Now searching " + layer.name +
                             "'s weight_fl...")

            metrics = []
            for test_fl in range(layer.bit_width):
                # save forward time
                if test_fl == layer.weight_fl:
                    metrics.append(metrics_tmp)
                else:
                    layer.trim_params(debug_bw=layer.bit_width,
                                      debug_fl=test_fl)
                    metrics_test = self.ext_metrics()
                    metrics.append(metrics_test)
                    # info log
                    self.logger.info("      weight_fl: " + str(test_fl) +
                                     ", metrics: " + str(metrics_test))

            # get maximum metrics' index
            layer.weight_fl = metrics.index(max(metrics))

            # info log
            self.logger.info("      " + layer.name + "'s weight_fl=" +
                             str(layer.weight_fl) + ", metrics=" +
                             str(max(metrics)))

            # maintain best weight_fl
            layer.trim_params(debug_bw=layer.bit_width,
                              debug_fl=layer.weight_fl)

            return max(metrics)

        def search_activation(layer, metrics_tmp):
            # info log
            self.logger.info("      Now searching " + layer.name +
                             "'s activation_fl...")

            metrics = []
            for test_fl in range(layer.bit_width):
                if test_fl == layer.activation_fl:
                    metrics.append(metrics_tmp)
                else:
                    layer.trim_activation(debug_bw=layer.bit_width,
                                          debug_fl=test_fl)
                    metrics_test = self.ext_metrics()
                    metrics.append(metrics_test)
                    # info log
                    self.logger.info("      activation_fl: " + str(test_fl) +
                                     ", metrics: " + str(metrics_test))
            # get maximum metrics' index
            layer.activation_fl = metrics.index(max(metrics))

            # info log
            self.logger.info("      " + layer.name + "'s activation_fl=" +
                             str(layer.activation_fl) + ", metrics=" +
                             str(max(metrics)))

            # maintain best activation_fl
            layer.trim_activation(debug_bw=layer.bit_width,
                                  debug_fl=layer.activation_fl)

            return max(metrics)

        def search_concat(layers):
            metrics = []
            # todo test all bit_width be same
            for test_fl in range(layers[0].bit_width):
                for layer in layers:
                    layer.trim_activation(debug_bw=layer.bit_width,
                                          debug_fl=test_fl)
                metrics_test = self.ext_metrics()
                metrics.append(metrics_test)
                # info log
                self.logger.info("      activation_fl: " + str(test_fl) +
                                 ", metrics: " + str(metrics_test))
            # get maximum metrics' index
            for layer in layers:
                layer.activation_fl = metrics.index(max(metrics))

            # maintain best activation_fl
            for layer in layers:
                layer.trim_activation(debug_bw=layer.bit_width,
                                      debug_fl=layer.activation_fl)

            return max(metrics), metrics.index(max(metrics))

        def do_quantize_params():
            # quantize params
            quantize_params()
            # test quantized metrics
            self.logger.info("  Testing quantized params metrics...")
            for layer in self.quant_layer_list:
                layer.trim_params()
            metrics_params = self.ext_metrics()
            self.logger.info("    Quantized params metrics:" +
                             str(metrics_params))

            return metrics_params, metrics_params

        def do_search_params(metrics_mid, metrics_params):
            #? only search when metrics drop 2 * self.threshold
            if metrics_mid - metrics_params >= 2 * self.search_threshold:
                # recover un-trim before search
                for layer in self.quant_layer_list:
                    layer.recover_params()
                # search params
                for index, layer in enumerate(self.quant_layer_list):
                    #? only search when layer have params
                    if layer.weight_fl != -1:
                        self.logger.info("  " + str(index + 1) + "/" +
                                         str(len(self.quant_layer_list)) +
                                         " Checking " + layer.name +
                                         "'s weight_fl.")
                        layer.trim_params()
                        metrics_tmp = self.ext_metrics()
                        #? only search when metrics drop >= self.threshold
                        if metrics_mid - metrics_tmp >= self.search_threshold:
                            self.logger.info("    Need to search, weight_fl=" +
                                             str(layer.weight_fl) +
                                             ", metrics=" + str(metrics_tmp))
                            metrics_mid = search_params(layer, metrics_tmp)
                        else:
                            metrics_mid = metrics_tmp
                            self.logger.info(
                                "    No need to search, weight_fl=" +
                                str(layer.weight_fl) + ", metrics=" +
                                str(metrics_mid))
            else:
                metrics_mid = metrics_params
                self.logger.info("    No need to search params, metrics=" +
                                 str(metrics_params))
            return metrics_mid, metrics_mid

        def do_quantize_activation():
            # process activation
            assert self.steper, "Search need a steper."
            self.logger.info("Gathering activation...")
            self.steper()
            for layer in self.quant_layer_list:
                layer.steper_sign = True
            print(" ")

            # quantize activation
            quantize_activation()
            # test quantized metrics
            self.logger.info("  Testing quantized activation metrics...")
            for layer in self.quant_layer_list:
                layer.trim_activation()
            metrics_activation = self.ext_metrics()
            self.logger.info("    Quantized activation metrics: " +
                             str(metrics_activation))
            return metrics_activation, metrics_activation

        def do_search_activation(metrics_mid, metrics_activation):
            #? only search only search when metrics drop 2 * self.threshold
            if metrics_mid - metrics_activation >= 2 * self.search_threshold:
                #todo valide
                #? remove handler
                # recovery un-trim before search
                for layer in self.quant_layer_list:
                    if layer.hook_handler is not None:
                        layer.hook_handler.remove()
                # search activation
                for index, layer in enumerate(self.quant_layer_list):
                    #? only search when layer need trim activation
                    if layer.activation_fl != -1:
                        self.logger.info("  " + str(index + 1) + "/" +
                                         str(len(self.quant_layer_list)) +
                                         " Checking " + layer.name +
                                         "'s activation_fl.")
                        layer.trim_activation()
                        metrics_tmp = self.ext_metrics()
                        #? only search when metrics drop >= self.threshold
                        if abs(metrics_tmp -
                               metrics_mid) >= self.search_threshold:
                            self.logger.info(
                                "    Need to search, activation_fl=" +
                                str(layer.activation_fl) + ", metrics=" +
                                str(metrics_tmp))
                            metrics_mid = search_activation(layer, metrics_tmp)
                        else:
                            metrics_mid = metrics_tmp
                            self.logger.info(
                                "    No Need to search, activation_fl=" +
                                str(layer.activation_fl) + ", metrics=" +
                                str(metrics_tmp))
            else:
                metrics_mid = metrics_activation
                self.logger.info("    No need to search activation, metrics=" +
                                 str(metrics_activation))
            return metrics_mid

        def do_search_concat(metrics_mid):
            if self.concat_num == 0:
                self.logger.warning(
                    "Concat_num is zero, no need to do_search_concat.")
                return metrics_mid
            else:
                metrics_cat = 0
            self.logger.info("There are " + str(self.concat_num) +
                             " concat(s) need to process.")
            # make every group concat fl same
            for index in range(self.concat_num):
                concat_layer_list = []
                for layer in self.quant_layer_list:
                    if layer.concat_sign == index:
                        concat_layer_list.append(layer)
                concat_layer_name = [x.name for x in concat_layer_list]
                self.logger.info("  Now processing the " + str(index + 1) +
                                 "th concat.")
                self.logger.info("  They are " + str(concat_layer_name))
                # These layers' bit_width should be same.
                # todo test bitwidth
                metrics_cat, group_fl = search_concat(concat_layer_list)
                self.logger.info("  " + str(index + 1) +
                                 "th processing done, group_fl=" +
                                 str(group_fl) + ", metrics=" +
                                 str(metrics_cat))
            return metrics_cat

        assert self.ext_metrics, "Search need a metrics."
        self.logger.info("Configuration: ")
        self.logger.info("  Log_path: " + str(self.log_path))
        self.logger.info("  Fuse_all: " + str(self.fuse_all))
        self.logger.info("  Concat_num: " + str(self.concat_num))
        self.logger.info("  Quantize_params: " + str(self.quantize_params))
        self.logger.info("  Search_params: " + str(self.search_params))
        self.logger.info("  Quantize_activation: " +
                         str(self.quantize_activation))
        self.logger.info("  Search_activation: " + str(self.search_activation))
        self.logger.info("  Search_concat: " + str(self.search_concat))
        self.logger.info("  Threshold: " + str(self.search_threshold))
        self.logger.info("  Testing origin metrics...")
        metrics_final = metrics_origin = metrics_mid = self.ext_metrics()
        self.logger.info("    Initialization metrics: " + str(metrics_mid))

        if self.quantize_params:
            metrics_final, metrics_params = do_quantize_params()
        if self.search_params:
            assert self.quantize_params, "If you want to search params, you must set quantize_params=true in config file."
            metrics_final, metrics_mid = do_search_params(
                metrics_mid, metrics_params)

        if self.quantize_activation:
            assert self.quantize_activation, "If you want to search activation, you must set quantize_activation=true in config file."
            metrics_final, metrics_activation = do_quantize_activation()
        if self.search_activation:
            metrics_final = do_search_activation(metrics_mid,
                                                 metrics_activation)

        if self.search_concat:
            metrics_final = do_search_concat(metrics_final)

        if not [
                self.quantize_params, self.search_params,
                self.quantize_activation, self.search_activation,
                self.search_concat
        ]:
            self.logger.warning(
                "Oops, we have done nothing, pls check the program, and you shouldn't set all false."
            )
        else:
            self.save_config()

            self.logger.info("Quantization finished: ")
            self.logger.info("  Origin metrics: " + str(metrics_origin))
            self.logger.info("  After metrics: " + str(metrics_final))

    def step(self):
        for layer in self.quant_layer_list:
            layer.init_histograms()
        #? simulate processing bar
        print(".", end=" ")

    def save_config(self, config_file=None):
        # save quant_config
        # quantization config and quantization parameters
        qconfig = {}

        # config
        qconfig['config'] = {}
        qconfig['config']['fuse_all'] = self.fuse_all
        qconfig['config']['concat_num'] = self.concat_num
        qconfig['config']['quantize_params'] = self.quantize_params
        qconfig['config']['search_params'] = self.search_params
        qconfig['config']['quantize_activation'] = self.quantize_activation
        qconfig['config']['search_activation'] = self.search_activation
        qconfig['config']['search_concat'] = self.search_concat
        qconfig['config']['search_threshold'] = self.search_threshold
        qconfig['config']['notes'] = "'-1' means no operation."

        # layers' quantization parameters
        for layer in self.quant_layer_list:
            qconfig[layer.name] = {}
            qconfig[layer.name]['bit_width'] = int(layer.bit_width)
            qconfig[layer.name]['weight_fl'] = int(layer.weight_fl)
            qconfig[layer.name]['activation_fl'] = int(layer.activation_fl)
            qconfig[layer.name]['concat_sign'] = int(layer.concat_sign)
            qconfig[layer.name]['scale_sign'] = layer.scale_sign
            qconfig[layer.name]['layer_type'] = str(type(layer.module))[8:-2]

        if config_file:
            with open(config_file, 'w+') as file:
                json.dump(qconfig, file)
        else:
            with open(self.log_path + '/quant.json', 'w+') as file:
                json.dump(qconfig, file)

    def load_config(self):
        # open config
        with open(self.config_file, 'r') as f:
            self.qconfig = json.load(f)
        # load config
        self.fuse_all = self.qconfig['config']['fuse_all']
        self.concat_num = self.qconfig['config']['concat_num']
        self.quantize_params = self.qconfig['config']['quantize_params']
        self.search_params = self.qconfig['config']['search_params']
        self.quantize_activation = self.qconfig['config'][
            'quantize_activation']
        self.search_activation = self.qconfig['config']['search_activation']
        self.search_concat = self.qconfig['config']['search_concat']
        self.search_threshold = self.qconfig['config']['search_threshold']

    def init_config(self):
        #! 1 same with self.load_config
        #! 2 same with self.save_config
        #! 3 same with config.json
        #! 3 same with logger info
        # set default config
        self.fuse_all = True
        self.concat_num = 0
        self.quantize_params = True
        self.search_params = True
        self.quantize_activation = True
        self.search_activation = True
        self.search_concat = True
        self.search_threshold = 0.005  # 0.5%
        self.bit_width = 8
        # modify model arch because fuse_all default is True
        fuse_module(self.model)
        # initialization
        self.quant_layer_list = []
        for name, module in self.model.named_modules():
            quant_layer = qnq_switch.get(type(module), None)
            if quant_layer:
                quant_layer = quant_layer(name, module, self.bit_width,
                                          self.logger, self.writer)
                self.quant_layer_list.append(quant_layer)
        # save init_config file
        self.save_config(self.config_file)

    def fuse(self):
        pass

    def trim_activation(self, index=None, debug_bw=None, debug_fl=None):
        self.logger.info("Turn on eval mode.")
        if index != None:
            self.quant_layer_list[index].trim_activation(debug_bw, debug_fl)
        else:
            for layer in self.quant_layer_list:
                layer.trim_activation()

    def transfer_bn2scale(self):
        for layer in self.quant_layer_list:
            if layer.scale_sign:
                bn2scale(layer.module)


# This script contain util will help you, good luck!
# Author:
#     Albert Dongz
# History:
#     2020.4.17 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing

import os
import torch
import logging
import datetime
import json
import torch.nn as nn


def lognow(path="./checkpoints/"):
    """This function set up logging.

    Keyword Arguments:
        path {str} -- log save path (default: {"./checkpoints/"})

    Returns:
        str -- log absolute path, contain time and param path.
    """
    # get current time and create log dir
    now = datetime.datetime.now().strftime('%Y-%m-%d--%H-%M-%S')
    log_path = os.path.normpath(path + '/' + now)
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    # setting up logging
    logger = logging.getLogger(__name__)
    logger.setLevel(level=logging.DEBUG)
    handler = logging.FileHandler(log_path + "/log.txt")
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.addHandler(console)
    logger.info('Start logging.')
    return logger, log_path


def float2fixed(data, bw, fl):
    """turn torch.Tensor to fixed Tensor.

    Args:
        data (torch.Tensor): torch.Tensor
        bw (int): bit width of quantization.
        fl (int): fraction length.

    Returns:
        torch.Tensor: fixed Tensor
    """
    unit = 2.0**(-fl)
    fixed_data = torch.clamp(torch.round(data / unit) * unit,
                             min=-2**(bw - 1 - fl) + unit,
                             max=+2**(bw - 1 - fl) - unit)
    return fixed_data


def bn2scale(module):
    """

    Args:
        module (nn.BatchNorm): Module need to be process
    """
    # 从未处理的module中获取bn层的四个参数
    weight = module.weight
    bias = module.bias
    try:
        mean = module.running_mean
    except:
        mean = 0
    try:
        var = module.running_var
    except:
        var = 1
    # 计算标准差，即sqrt(var) => std
    std = torch.sqrt(var)
    # 计算值
    new_weight = weight / std
    new_bias = bias - weight * mean / std
    new_mean = torch.zeros(mean.shape)
    new_var = torch.ones(var.shape)
    # 保存参数
    module.weight = nn.Parameter(new_weight)
    module.bias = nn.Parameter(new_bias)
    module.running_mean = new_mean
    module.running_var = new_var


def transfer_bn2scale(net):
    """Auto transfer BatchNorm to Scale

    Arguments:
        net {torch.nn.Module} -- net need be processed
        state {OrderDict} -- checkpoint of net

    Returns:
        net, state -- processed net and state
    """
    for name, module in net.named_modules():
        if isinstance(module, nn.BatchNorm2d):
            bn2scale(module)
    # todo
    net.cuda()


class DummyModule(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


# Ref: https://zhuanlan.zhihu.com/p/49329030
# fuse Conv-BN & Deconv-BN
def fuse_module(m):
    def fuse(conv, bn):
        w = conv.weight
        mean = bn.running_mean
        var_sqrt = torch.sqrt(bn.running_var + bn.eps)

        beta = bn.weight
        gamma = bn.bias

        if conv.bias is not None:
            b = conv.bias
        else:
            b = mean.new_zeros(mean.shape)

        w = w * (beta / var_sqrt).reshape([conv.out_channels, 1, 1, 1])
        b = (b - mean) / var_sqrt * beta + gamma
        if isinstance(conv, nn.Conv2d):
            fused_conv = nn.Conv2d(conv.in_channels,
                                   conv.out_channels,
                                   conv.kernel_size,
                                   conv.stride,
                                   conv.padding,
                                   bias=True)
        else:
            fused_conv = nn.ConvTranspose2d(conv.in_channels,
                                            conv.out_channels,
                                            conv.kernel_size,
                                            conv.stride,
                                            conv.padding,
                                            bias=True)
        fused_conv.weight = nn.Parameter(w)
        fused_conv.bias = nn.Parameter(b)
        return fused_conv

    children = list(m.named_children())
    c = None
    cn = None

    for name, child in children:
        if isinstance(child, nn.BatchNorm2d):
            bc = fuse(c, child)
            m._modules[cn] = bc
            m._modules[name] = DummyModule()
            c = None
        elif isinstance(child, (nn.Conv2d, nn.ConvTranspose2d)):
            c = child
            cn = name
        else:
            fuse_module(child)


def fuse_convbn(model, conv_layer, bn_layer):
    conv = conv_layer.module
    bn = bn_layer.module

    # BN = bn_alpha * ( (x - bn_mean) / bn_std ) + bn_beta
    bn_mean = bn.running_mean
    bn_std = torch.sqrt(bn.running_var + bn.eps)
    bn_alpha = bn.weight
    bn_beta = bn.bias
    # Conv = conv_weight * x + conv_bias
    conv_weight = conv.weight
    conv_bias = conv.bias if conv.bias is not None else bn_mean.new_zeros(
        bn_mean.shape)

    # fused parameters
    fused_weight = conv_weight * (bn_alpha / bn_std).reshape(
        [conv.out_channels, 1, 1, 1])
    fused_bias = (conv_bias - bn_mean) / bn_std * bn_alpha + bn_beta

    # fuse
    fused_conv = nn.Conv2d(conv.in_channels,
                           conv.out_channels,
                           conv.kernel_size,
                           conv.stride,
                           conv.padding,
                           bias=True)
    fused_conv.weight = nn.Parameter(fused_weight)
    fused_conv.bias = nn.Parameter(fused_bias)

    model[conv_layer.name] = fused_conv
    model[bn_layer.name] = DummyModule()


def fuse_convbnrelu(model, conv_layer, bn_layer, relu_layer):
    # todo add relu
    conv = conv_layer.module
    bn = bn_layer.module

    # BN = bn_alpha * ( (x - bn_mean) / bn_std ) + bn_beta
    bn_mean = bn.running_mean
    bn_std = torch.sqrt(bn.running_var + bn.eps)
    bn_alpha = bn.weight
    bn_beta = bn.bias
    # Conv = conv_weight * x + conv_bias
    conv_weight = conv.weight
    conv_bias = conv.bias if conv.bias is not None else bn_mean.new_zeros(
        bn_mean.shape)

    # fused parameters
    fused_weight = conv_weight * (bn_alpha / bn_std).reshape(
        [conv.out_channels, 1, 1, 1])
    fused_bias = (conv_bias - bn_mean) / bn_std * bn_alpha + bn_beta

    # fuse
    fused_conv = nn.Conv2d(conv.in_channels,
                           conv.out_channels,
                           conv.kernel_size,
                           conv.stride,
                           conv.padding,
                           bias=True)
    fused_conv.weight = nn.Parameter(fused_weight)
    fused_conv.bias = nn.Parameter(fused_bias)

    model[conv_layer.name] = fused_conv
    model[bn_layer.name] = DummyModule()


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from torch.nn.modules.sparse import Embedding
from .quant_base import QuantBase


class QuantUpsample(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False

    def quantize_params(self):
        pass


class QuantEmbedding(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False


vision_switch = {nn.Upsample: QuantUpsample, nn.Embedding: QuantEmbedding}



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing

import torch.nn as nn


class QNQInput(nn.Module):
    def __init__(self):
        super(QNQInput, self).__init__()

    def forward(self, x):
        return x


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from torch.nn.modules.activation import LeakyReLU, PReLU
from .quant_base import QuantBase


class QuantRelu(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = True
        self.weight_fl = -1

    def quantize_params(self):
        pass


class QuantLRelu(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False
        self.weight_fl = -1

    def quantize_params(self):
        pass


class QuantPRelu(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False
        self.weight_fl = -1

    def quantize_params(self):
        pass


class QuantLogSoftmax(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False
        self.weight_fl = -1

    def quantize_params(self):
        pass


activation_switch = {
    # nn.ReLU: QuantRelu,
    # nn.LeakyReLU: QuantLRelu,
    # nn.PReLU: QuantPRelu,
    nn.LogSoftmax:
    QuantLogSoftmax
}


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from .quant_base import QuantBase

sparse_switch = {}


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch
import torch.nn as nn


class TorchAdd(nn.Module):
    def __init__(self):
        super(TorchAdd, self).__init__()

    def forward(self, x, y):
        return x + y


class TorchSum(nn.Module):
    def __init__(self):
        super(TorchSum, self).__init__()

    def forward(self, *args, **kwargs):
        return torch.sum(*args, **kwargs)


class TorchMinus(nn.Module):
    def __init__(self):
        super(TorchMinus, self).__init__()

    def forward(self, x, y):
        return x - y


class TorchMatMul(nn.Module):
    def __init__(self):
        super(TorchMatMul, self).__init__()

    def forward(self, x, y):
        return torch.matmul(x, y)


class TorchDotMul(nn.Module):
    def __init__(self):
        super(TorchDotMul, self).__init__()

    def forward(self, x, y):
        return x * y


class TorchDiv(nn.Module):
    def __init__(self):
        super(TorchDiv, self).__init__()

    def forward(self, x, y):
        return x / y


class TorchSqrt(nn.Module):
    def __init__(self):
        super(TorchSqrt, self).__init__()

    def forward(self, x):
        return torch.sqrt(x)


class TorchSin(nn.Module):
    def __init__(self):
        super(TorchSin, self).__init__()

    def forward(self, x):
        return torch.sin(x)


class TorchCos(nn.Module):
    def __init__(self):
        super(TorchCos, self).__init__()

    def forward(self, x):
        return torch.cos(x)


class TorchSoftmax(nn.Module):
    def __init__(self):
        super(TorchSoftmax, self).__init__()

    def forward(self, *args, **kwargs):
        return torch.softmax(*args, **kwargs)


class TorchSigmoid(nn.Module):
    def __init__(self):
        super(TorchSigmoid, self).__init__()

    def forward(self, x):
        return torch.sigmoid(x)


class TorchExp(nn.Module):
    def __init__(self):
        super(TorchExp, self).__init__()

    def forward(self, x):
        return torch.exp(x)


class TorchTanh(nn.Module):
    def __init__(self):
        super(TorchTanh, self).__init__()

    def forward(self, x):
        return torch.tanh(x)


class TorchTemplate(nn.Module):
    def __init__(self, func):
        super(TorchTemplate, self).__init__()
        self.func = func

    def forward(self, *args):
        return self.func(*args)


class TorchDummy(nn.Module):
    """TorchDummy collect activation from different input and quantize these use unified fl choice.

    Args:
        x (any): You can use any algorithm but a single function,
        eg: 'a + b' or 'a * b' or just a single variable 'a',
        but not 'a * (b + c)', that's because (b + c) in later need be quantized too.
    """
    def __init__(self):
        super(TorchDummy, self).__init__()

    def forward(self, x):
        return x


class TorchCat(nn.Module):
    def __init__(self, *layers):
        super(TorchCat, self).__init__()
        self.layers = layers

    def forward(self, *args):
        return torch.cat(*args)


class ConcatSign(nn.Module):
    def __init__(self, modules=[]):
        super(ConcatSign, self).__init__()
        self.modules = modules

    def forward(self):
        pass


# global var
TorchFuncTuple = (TorchAdd, TorchSum, TorchMinus, TorchMatMul, TorchDotMul,
                  TorchDiv, TorchSqrt, TorchSin, TorchCos, TorchSoftmax,
                  TorchSigmoid, TorchExp, TorchTanh, TorchTemplate, TorchDummy)



import json
# This script contain func help generating a onnx-caffe index.
# Author:
#     Albert Dongz
# History:
#     2020.10.15 First Release
# Dependencies:
#     json
# Attention:
#     1. Nothing


def gen_index(json_file, onnx_model):
    """generate a onnx-caffe index

    Args:
        json_file (str): json file you want to save
        onnx_model (onnx_model): model load from onnx file
    """
    with open(json_file, 'w+') as f:
        index = list(
            map(
                lambda x: {
                    'operator': str(x.op_type),
                    'layer_num': int(list(x.output)[0]),
                    'param': list(x.input)
                }, list(onnx_model.graph.node)))
        json.dump(index, f)



# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from torch.nn.modules.pooling import AdaptiveAvgPool1d
from .quant_base import QuantBase


class QuantMaxPool(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False

    def quantize_params(self):
        pass


class QuantAveragePool(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False

    def quantize_params(self):
        pass


class QuantAdaptiveAvgPool(QuantBase):
    def __init__(self, name, module, bit_width, is_debug=False):
        super().__init__(name, module, bit_width, is_debug)
        self.activation_positive = False

    def quantize_params(self):
        pass


pooling_switch = {
    # nn.MaxPool1d: QuantMaxPool,
    # nn.MaxPool2d: QuantMaxPool,
    # nn.MaxPool3d: QuantMaxPool,
    nn.AvgPool1d:
    QuantAveragePool,
    nn.AvgPool2d:
    QuantAveragePool,
    nn.AvgPool3d:
    QuantAveragePool,
    nn.AdaptiveAvgPool1d:
    QuantAdaptiveAvgPool,
    nn.AdaptiveAvgPool2d:
    QuantAdaptiveAvgPool,
    nn.AdaptiveAvgPool3d:
    QuantAdaptiveAvgPool,
}


# This script contain decorators help developing.
# Author:
#     Albert Dongz
# History:
#     2020.4.17 First Release
# Dependencies:
#     time
# Attention:
#     1. Nothing

import time


def exe_time(func):
    """
    This function is a wrapper which can measure the exec time.

    Arguments:
        func {function} -- func need to be measure.

    Returns:
        function -- a func can measure exec time.
    """
    def new_func(*args, **args2):
        t0 = time.time()
        print("@%s, {%s} start" %
              (time.strftime("%X", time.localtime()), func.__name__))
        back = func(*args, **args2)
        print("@%s, {%s} end" %
              (time.strftime("%X", time.localtime()), func.__name__))
        print("@%.3fs taken for {%s}" % (time.time() - t0, func.__name__))
        return back

    return new_func


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch
import torch.nn as nn
import numpy as np
from torch.nn.modules import normalization
from .quant_base import *


class QuantBatchNorm(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False


class QuantLayerNorm(QuantBase):
    def __init__(self, name, module, bit_width, is_debug):
        super().__init__(name, module, bit_width, is_debug)


normalization_switch = {
    nn.BatchNorm1d: QuantBatchNorm,
    nn.BatchNorm2d: QuantBatchNorm,
    nn.BatchNorm3d: QuantBatchNorm,
    nn.LayerNorm: QuantLayerNorm
}


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
import numpy as np
from .quant_base import QuantBase
from ..qnq_ops import *


class QuantTorchFuncGeneral(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False
        self.weight_fl = -1

    def quantize_params(self):
        pass


torchfunc_switch = {
    TorchAdd: QuantTorchFuncGeneral,
    TorchSum: QuantTorchFuncGeneral,
    TorchMinus: QuantTorchFuncGeneral,
    TorchMatMul: QuantTorchFuncGeneral,
    TorchDotMul: QuantTorchFuncGeneral,
    TorchDiv: QuantTorchFuncGeneral,
    TorchSqrt: QuantTorchFuncGeneral,
    TorchSin: QuantTorchFuncGeneral,
    TorchCos: QuantTorchFuncGeneral,
    TorchSoftmax: QuantTorchFuncGeneral,
    TorchSigmoid: QuantTorchFuncGeneral,
    TorchExp: QuantTorchFuncGeneral,
    TorchTanh: QuantTorchFuncGeneral,
    TorchTemplate: QuantTorchFuncGeneral,
    TorchDummy: QuantTorchFuncGeneral,
}


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from ..qnq_ops import *
from .quant_base import *


class QuantLSTM(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_fl = -1

    def quantize_params(self):
        # overide backup data
        self.backup_params = {
            k: v
            for k, v in zip(self.module.params, [
                copy.deepcopy(getattr(self.module, name))
                for name in self.module.params
            ])
        }

        for k, v in self.backup_params.items():
            self.writer.add_histogram(self.name + k,
                                      v.clone().cpu().data.numpy())

        #! TODO valide
        params = np.concatenate([
            getattr(self.module, attr).cpu().data.flatten().numpy()
            for attr in self.module.params
        ])
        # init kld
        hist, _ = np.histogram(params,
                               bins=self.bins_num,
                               range=(-1 * self.activation_max,
                                      self.activation_max))
        min_kld_fl, kld = kld_based_dynamics_double(hist, self.bit_width)
        self.weight_fl = min_kld_fl
        self.logger.debug(self.name + 's\' weight fl:' + str(min_kld_fl) +
                          ', kld: ' + str(kld))

    def trim_params(self, debug_bw=None, debug_fl=None):
        # for debug
        bit_width = self.bit_width if not debug_bw else debug_bw
        fraction_length = self.weight_fl if not debug_fl else debug_fl

        # trim from backed data
        for name in self.module.params:
            getattr(self.module,
                    name).data = float2fixed(self.backup_params[name].data,
                                             bit_width, fraction_length)

    def recover_params(self):
        # recover from backed data
        for name in self.module.params:
            getattr(self.module, name).data = self.backup_params[name].data

    def quantize_activation(self):
        pass

    def register_eval_hook(self, debug_bw=None, debug_fl=None):
        pass

    def register_get_activation_hook(self):
        pass

    def trim_activation(self, debug_bw=None, debug_fl=None):
        pass

    def init_histograms(self):
        pass


class QuantGRU(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_fl = -1

    def quantize_params(self):
        #! TODO valide
        params = np.concatenate([
            getattr(self.module, attr).cpu().data.flatten().numpy()
            for attr in self.module.params
        ])
        # init kld
        hist, _ = np.histogram(params,
                               bins=self.bins_num,
                               range=(-1 * self.activation_max,
                                      self.activation_max))
        min_kld_fl, kld = kld_based_dynamics_double(hist, self.bit_width)
        self.weight_fl = min_kld_fl
        for name in self.module.params:
            getattr(self.module, name).data = float2fixed(
                getattr(self.module, name).data, self.bit_width,
                self.weight_fl)
        self.logger.debug(self.name + 's\' weight fl:' + str(min_kld_fl) +
                          ', kld: ' + str(kld))

    def quantize_activation(self):
        pass

    def register_eval_hook(self, debug_bw=None, debug_fl=None):
        pass

    def register_get_activation_hook(self):
        pass

    def trim_activation(self, debug_bw=None, debug_fl=None):
        pass

    def init_histograms(self):
        pass


recurrent_switch = {
    QNQLSTM: QuantLSTM,
    QNQGRU: QuantGRU,
}


# This script contain implentation of operators.
# Author:
#     Albert Dongz
# History:
#     2020.7.16 First Release
# Dependencies:
#     PyTorch
# Attention:
#     1. Nothing
import torch.nn as nn
from .quant_base import QuantBase


class QuantLinear(QuantBase):
    def __init__(self, name, module, bit_width, logger, writer):
        super().__init__(name, module, bit_width, logger, writer)
        self.activation_positive = False


linear_switch = {nn.Linear: QuantLinear}
